/*
 *  Copyright (c) 2024 John Ahlroos
 *
 *  Creative Commons Attribution-NoDerivatives 4.0 International Public License
 *
 *  By exercising the Licensed Rights (defined below), You accept and agree
 *  to be bound by the terms and conditions of this Creative Commons
 *  Attribution-NoDerivatives 4.0 International Public License ("Public
 *  License"). To the extent this Public License may be interpreted as a
 *  contract, You are granted the Licensed Rights in consideration of Your
 *  acceptance of these terms and conditions, and the Licensor grants You
 *  such rights in consideration of benefits the Licensor receives from
 *  making the Licensed Material available under these terms and
 *  conditions.
 *
 *  Refer to LICENSE file or https://creativecommons.org/licenses/by-nd/4.0 for full license.
 *
 */
package com.devsoap.json;

import jakarta.json.*;
import jakarta.json.stream.JsonParser;
import org.assertj.core.api.AbstractAssert;
import org.assertj.core.api.ListAssert;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;

public class JsonAssert extends AbstractAssert<JsonAssert, String> {

    protected JsonAssert(String json) {
        super(json, JsonAssert.class);
    }

    public static JsonAssert assertThat(String json) {
        return new JsonAssert(json);
    }
    public static JsonAssert assertThat(Object object) {
        return assertThat(object.toString());
    }

    public JsonValueAssert<?> firstField(String fieldName) {
        try (var stream = new ByteArrayInputStream(actual.getBytes(StandardCharsets.UTF_8))) {
            var actualParser = Json.createParser(stream);
            actualParser.next(); // START_OBJECT
            return findFieldInObjectOnAnyDepth(fieldName, actualParser)
                    .orElseThrow(() -> failure("Field %s not found.", fieldName) );
        } catch (IOException e) {
            throw failure("Failed to parse field");
        }
    }

    public JsonValueAssert<?> lastField(String fieldName) {
        try (var stream = new ByteArrayInputStream(actual.getBytes(StandardCharsets.UTF_8))) {
            var scannerParser = Json.createParser(stream);
            scannerParser.next(); // START_OBJECT
            return findLastFieldInObject(fieldName, scannerParser)
                    .orElseThrow(() -> failure("Field %s not found.", fieldName) );
        } catch (IOException e) {
            throw failure("Failed to parse field");
        }
    }

    public JsonValueAssert<?> nthField(String fieldName, int index) {
        try (var stream = new ByteArrayInputStream(actual.getBytes(StandardCharsets.UTF_8))) {
            var parser = Json.createParser(stream);
            parser.next(); // START_OBJECT
            return findNthFieldInObject(fieldName, parser, index)
                    .orElseThrow(() -> failure("%d nth field %s not found.", index, fieldName) );
        } catch (IOException e) {
            throw failure("Failed to parse field");
        }
    }

    public JsonValueAssert<?> field(String fieldName) {
        try (var stream = new ByteArrayInputStream(actual.getBytes(StandardCharsets.UTF_8))) {
            var actualParser = Json.createParser(stream);
            actualParser.next(); // START_OBJECT
            return findFieldInObject(fieldName, actualParser)
                    .orElseThrow(() -> failure("Field %s not found.", fieldName) );
        } catch (IOException e) {
            throw failure("Failed to parse field");
        }
    }

    public JsonArrayAssert<?> array(String fieldName) {
        try (var stream = new ByteArrayInputStream(actual.getBytes(StandardCharsets.UTF_8))) {
            var actualParser = Json.createParser(stream);
            actualParser.next(); // START_OBJECT
            return findArrayInObject(fieldName, actualParser);
        } catch (IOException e) {
            throw failure("Failed to parse field");
        }
    }

    public <ITEM> ListAssert<ITEM> array(String fieldName, Class<ITEM> itemClass) {
        try (var stream = new ByteArrayInputStream(actual.getBytes(StandardCharsets.UTF_8))) {
            var actualParser = Json.createParser(stream);
            actualParser.next(); // START_OBJECT
            return findArrayInObject(fieldName, actualParser, itemClass);
        } catch (IOException e) {
            throw failure("Failed to parse field");
        }
    }

    public JsonValueAssert<?> path(String... fields) {
        var actualParser = Json.createParser(new ByteArrayInputStream(actual.getBytes(StandardCharsets.UTF_8)));
        actualParser.next(); // START_OBJECT
        var fieldList = new ArrayList<>(List.of(fields));
        var currentField = field(fieldList.removeFirst());
        return currentField.path(fieldList.toArray(String[]::new));
    }

    private static Optional<JsonValueAssert<?>> findFieldInObject(String fieldName, JsonParser parser) {
        if (parser.currentEvent() != JsonParser.Event.START_OBJECT) {
            throw new IllegalArgumentException("Current event must be START_OBJECT");
        }
        return typedJsonValue(fieldName, parser.getObject().get(fieldName));
    }

    private static JsonArrayAssert<?> findArrayInObject(String fieldName, JsonParser parser) {
        if (parser.currentEvent() != JsonParser.Event.START_OBJECT) {
            throw new IllegalArgumentException("Current event must be START_OBJECT");
        }
        return typedJsonArray(fieldName, parser.getObject().get(fieldName));
    }

    private static <ITEM> ListAssert<ITEM> findArrayInObject(String fieldName, JsonParser parser, Class<ITEM> itemClass) {
        if (parser.currentEvent() != JsonParser.Event.START_OBJECT) {
            throw new IllegalArgumentException("Current event must be START_OBJECT");
        }
        return typedJsonArray(parser.getObject().get(fieldName));
    }

    private static Optional<JsonValueAssert<?>> findFieldInObjectOnAnyDepth(String fieldName, JsonParser parser) {
        if (parser.currentEvent() != JsonParser.Event.START_OBJECT) {
            throw new IllegalArgumentException("Current event must be START_OBJECT");
        }
        while (parser.hasNext()) {
            var keyEvent = parser.next();
            if (keyEvent == JsonParser.Event.KEY_NAME ) {
                var key = parser.getString();
                parser.next(); // VALUE
                if (key.equalsIgnoreCase(fieldName)) {
                    return typedJsonValue(key, parser.getValue());
                }
            }
        }
        return Optional.empty();
    }

    private static Optional<JsonValueAssert<?>> findNthFieldInObject(String fieldName, JsonParser parser, int index) {
        if (parser.currentEvent() != JsonParser.Event.START_OBJECT) {
            throw new IllegalArgumentException("Current event must be START_OBJECT");
        }
        var counter = 0;
        while (parser.hasNext()) {
            var keyEvent = parser.next();
            if (keyEvent == JsonParser.Event.KEY_NAME ) {
                var key = parser.getString();
                parser.next(); // VALUE
                if (key.equalsIgnoreCase(fieldName)) {
                    if (counter == index) {
                        return typedJsonValue(key, parser.getValue());
                    }
                    counter++;
                }
            }
        }
        return Optional.empty();
    }

    private static Optional<JsonValueAssert<?>> findLastFieldInObject(String fieldName, JsonParser scannerParser) {
        if (scannerParser.currentEvent() != JsonParser.Event.START_OBJECT) {
            throw new IllegalArgumentException("Current event must be START_OBJECT");
        }

        JsonValue current = null;
        while (scannerParser.hasNext()) {
            var keyEvent = scannerParser.next();
            if (keyEvent == JsonParser.Event.KEY_NAME ) {
                var key = scannerParser.getString();
                scannerParser.next(); // VALUE
                if (key.equalsIgnoreCase(fieldName)) {
                    current = scannerParser.getValue();
                }
            }
        }

        if (current == null) {
            return Optional.empty();
        }

        if (current.getValueType() == JsonValue.ValueType.OBJECT) {
            try (var stream = new ByteArrayInputStream(current.toString().getBytes(StandardCharsets.UTF_8))) {
                var parser = Json.createParser(stream);
                parser.next(); // START_OBJECT
                var found = findLastFieldInObject(fieldName, parser);
                if (found.isPresent()) {
                    return found;
                }
            } catch (IOException e) {
                throw new IllegalStateException("Failed to parse json value");
            }
        }

        return Optional.of(current).flatMap(v -> typedJsonValue(fieldName, v));
    }


    static Optional<JsonValueAssert<?>> typedJsonValue(String fieldName, JsonValue jsonValue) {
        return Optional.ofNullable(jsonValue).map(v -> switch (jsonValue.getValueType()) {
            case ARRAY -> null;
            case OBJECT -> new JsonValueAssert<>(fieldName, jsonValue, jsonValue.asJsonObject());
            case STRING -> new JsonValueAssert<>(fieldName, jsonValue, ((JsonString) jsonValue).getString());
            case TRUE -> new JsonValueAssert<>(fieldName, jsonValue, true);
            case FALSE -> new JsonValueAssert<>(fieldName, jsonValue, false);
            case NULL -> new JsonValueAssert<>(fieldName, jsonValue, null);
            case NUMBER -> {
                Number number = ((JsonNumber) jsonValue).numberValue();
                if (number instanceof Integer) {
                    yield new JsonValueAssert<>(fieldName, jsonValue, (Integer) number);
                }
                if (number instanceof Short) {
                    yield new JsonValueAssert<>(fieldName, jsonValue, (Short) number);
                }
                if (number instanceof Long) {
                    yield new JsonValueAssert<>(fieldName, jsonValue, (Long) number);
                }
                if (number instanceof Double) {
                    yield new JsonValueAssert<>(fieldName, jsonValue, (Double) number);
                }
                if (number instanceof Float) {
                    yield new JsonValueAssert<>(fieldName, jsonValue, (Float) number);
                }
                if (number instanceof BigDecimal) {
                    yield new JsonValueAssert<>(fieldName, jsonValue, (BigDecimal) number);
                }
                if (number instanceof BigInteger) {
                    yield new JsonValueAssert<>(fieldName, jsonValue, (BigInteger) number);
                }
                yield new JsonValueAssert<>(fieldName, jsonValue, ((JsonNumber) jsonValue).bigDecimalValue());
            }
        });
    }

    static JsonArrayAssert<?> typedJsonArray(String fieldName, JsonValue jsonArray) {
        return new JsonArrayAssert<>(fieldName, (JsonArray) jsonArray);
    }
    static <ITEM> ListAssert<ITEM> typedJsonArray(JsonValue jsonArray) {
        return new ListAssert<>(((JsonArray)jsonArray).getValuesAs(value -> (ITEM) switch (value.getValueType()) {
            case ARRAY -> null;
            case OBJECT -> null;
            case STRING -> ((JsonString) value).toString();
            case NUMBER -> ((JsonNumber) value).intValue();
            case TRUE -> Boolean.TRUE;
            case FALSE -> Boolean.FALSE;
            case NULL -> null;
        }));
    }
}